from json import dumps
from typing import Tuple
from django import forms
from django.conf import settings
from django.contrib.admin import widgets
from json import dumps
# from django.contrib import admin
from django.urls import reverse


class ExtendedAutocompleteSelect(widgets.AutocompleteSelect):

    template_name = 'djeu/forms/widgets/select.html'
    option_template_name = 'djeu/forms/widgets/select_option.html'

    # FIXME: https://stackoverflow.com/questions/63199404/django-choice-list-dynamic-choices
    def __init__(self, *args, admin_site=None, url_name=None, use_admin_site_jquery=False, additional_data_components=None, **kwargs):
        super(ExtendedAutocompleteSelect, self).__init__(*args, admin_site, **kwargs)
        self.allow_multiple_selected = False
        self.additional_data_components = additional_data_components
        if not hasattr(self, 'i18n_name'):
            self.i18n_name = None
        self.use_admin_site_jquery = use_admin_site_jquery
        if url_name:
            self.url_name = url_name

    def _label(self, obj) -> str:
        return str(obj)

    def optgroups(self, name, value, attr=None):
        """Return selected options based on the ModelChoiceIterator."""
        default = (None, [], 0)
        groups = [default]
        has_selected = False
        selected_choices = {
            str(v) for v in value
            if str(v) not in self.choices.field.empty_values
        }
        if not self.is_required and not self.allow_multiple_selected:
            default[1].append(self.create_option(name, '', '', False, 0))
        remote_model_opts = self.field.remote_field.model._meta
        to_field_name = getattr(
            self.field.remote_field, 'field_name', remote_model_opts.pk.attname)
        to_field_name = remote_model_opts.get_field(to_field_name).attname

        choices = (
            (getattr(obj, to_field_name),
                self.choices.field.label_from_instance(obj))
            for obj in self.choices.queryset.using(self.db).filter(**{'%s__in' % to_field_name: selected_choices})
        )

        for option_value, option_label in choices:
            selected = (
                str(option_value) in value and
                (has_selected is False or self.allow_multiple_selected)
            )
            has_selected |= selected
            index = len(default[1])
            subgroup = default[1]
            subgroup.append(self.create_option(
                name, option_value, option_label, selected_choices, index))

        return groups

    def create_option(self, name, value, label, selected, index, subindex=None, attrs=None):
        result = super(ExtendedAutocompleteSelect, self).create_option(
            name, value, label, selected, index, subindex, attrs)
        if attrs:
            result['attrs'] = {**result['attrs'], **attrs}
        return result
    
    def get_url(self):
        if self.admin_site:
            return super(ExtendedAutocompleteSelect, self).get_url()
        return reverse(self.url_name)

    def get_context(self, name, value, attrs):
        # Request models and fields for the autocomplete requests
        # As an example:
        #
        # For the many to many field in document
        # through_fields = (document, relative, kind)
        # => relative: relationship
        # => kind: relationship
        extended_foreign_key_field = self.field
        relation_model_name = extended_foreign_key_field.model._meta.model_name
        # owner_model_name = extended_foreign_key_field.model._meta.model_name

        data_components = [{field: relation_model_name}
                           for field in extended_foreign_key_field.key_data_components]
        if self.additional_data_components:
            for additional_data_component in self.additional_data_components:
                if additional_data_component in data_components:
                    data_components.remove(additional_data_component)
            data_components.extend(self.additional_data_components)

        result = super(ExtendedAutocompleteSelect, self).get_context(name, value, {**attrs, **{
            'data-components': dumps(data_components),
            'data-data': dumps([]),
            'data-tags': True,
        }})
        # remove garbage
        widget_attrs = result['widget']['attrs']
        # if 'data-model-name' in widget_attrs:
        #     del widget_attrs['data-model-name']
        # if 'data-field-name' in widget_attrs:
        #     del widget_attrs['data-field-name']
        return result

    @property
    def media(self):
        extra = '' if settings.DEBUG else '.min'
        i18n_file = ('admin/js/vendor/select2/i18n/%s.js' %
                     self.i18n_name,) if self.i18n_name else ()
        if self.use_admin_site_jquery:
            return forms.Media(
                js=(
                    'admin/js/vendor/jquery/jquery%s.js' % extra,
                    'djeu/js/jquery-extended-autocomplete-select-multiply.js',
                ) + i18n_file + (
                    'admin/js/jquery.init.js',
                    'djeu/js/jquery-extended-autocomplete-select-multiply.init.js',
                ),
                css={
                    'screen': (
                        'djeu/css/jquery-extended-autocomplete-select-multiply.css',
                        'djeu/css/autocomplete.mod.css',
                    ),
                },
            )
        else:
            return forms.Media(
                js=(
                    'djeu/js/jquery-extended-autocomplete-select-multiply.js',
                ) + i18n_file + (
                    'djeu/js/jquery-extended-autocomplete-select-multiply.init.js',
                ),
                css={
                    'screen': (
                        'djeu/css/jquery-extended-autocomplete-select-multiply.css',
                        'djeu/css/autocomplete.mod.css',
                    ),
                },
            )



from re import split as regexp_split
from typing import Optional

from django.db import models
from django.db.models import Manager
from django.contrib import admin
from django.forms.models import ModelMultipleChoiceField

from .widgets.extended_autocomplete_select_multiply import ExtendedAutocompleteSelectMultiple


class ExtendedModelMultipleChoiceField(ModelMultipleChoiceField):

    widget = ExtendedAutocompleteSelectMultiple

    def __init__(self, queryset, **kwargs):
        if 'widget' in kwargs:
            widget = kwargs.pop('widget')
            kwargs.pop('django_field')
        else:
            widget = self.widget(kwargs.pop('django_field'), admin.site)
        super().__init__(queryset,
                         widget=widget,
                         **kwargs)
        self.missed_keys_cache = dict()

    def label_from_instance(self, obj):
        """
        Convert objects into strings and generate the labels for the choices
        presented by this object. Subclasses can override this method to
        customize the display of the choices.
        """
        return str(obj)

    def to_python(self, value):
        return value

    def _default_procedure_to_missed_key(self, django_model, particular_value) -> Optional[int]:
        missed_key_key = f'{django_model._meta.model_name}.{particular_value}'.lower(
        )
        if missed_key_key in self.missed_keys_cache:
            return self.missed_keys_cache[missed_key_key]

        instance = django_model()
        tokens = regexp_split(r'\s+', particular_value)
        char_fields = [field for field in django_model._meta.fields if isinstance(
            field, models.CharField)]
        if len(tokens) <= len(char_fields):
            for i in range(0, len(tokens)):
                setattr(instance, char_fields[i].attname, tokens[i])
        elif len(char_fields) > 0:
            setattr(instance, char_fields[0].name, particular_value)
        else:
            raise BaseException(
                'Can\'t set automatically {particular_value} for {django_model}. Try to do in manual manner.')

        instance.save()
        self.missed_keys_cache[missed_key_key] = instance.pk
        return instance.pk

    def _compound_key_values(self, through, compound_key, additional_fields,
                             values, missed_key_procedure):
        result_count = 0
        result = None
        for value in values:  # value is comma separated string
            compound_key_values = dict()
            previous_comma_position = 0
            comma_position = str(value).find(',')
            for particular_field_name in compound_key:
                particular_field_value = str(value)[
                    previous_comma_position:comma_position] if comma_position > -1 else str(value)[previous_comma_position:]
                pk_value = int(
                    particular_field_value) if particular_field_value.isdigit() else None
                if not pk_value and missed_key_procedure:
                    # For ForeignKey
                    django_model = through._meta.get_field(
                        particular_field_name).related_model if self.widget.field.through else through
                    pk_value = missed_key_procedure(
                        django_model, particular_field_value)
                if not pk_value:
                    break
                compound_key_values[particular_field_name] = pk_value

                previous_comma_position = comma_position + 1 if comma_position > -1 else -1
                if previous_comma_position > -1:
                    comma_position = str(value).find(
                        ',', previous_comma_position)

            additional_fields_values = dict()
            for additional_field_name in additional_fields:
                if previous_comma_position == -1 and comma_position == -1:
                    additional_fields_values[additional_field_name] = ''
                    continue
                particular_field_value = str(value)[previous_comma_position:] if comma_position == -1 else str(
                    value)[previous_comma_position:comma_position]
                additional_fields_values[additional_field_name] = particular_field_value

                previous_comma_position = comma_position + 1 if comma_position > -1 else -1
                if previous_comma_position > -1:
                    comma_position = str(value).find(
                        ',', previous_comma_position)

            # check full key
            if len(compound_key_values) == len(compound_key):
                stored_values = {**compound_key_values,
                                 **additional_fields_values}
                result_count += 1
                if result_count == 1:
                    result = stored_values
                elif result_count == 2:
                    for particular_field_name in [*compound_key_values, *additional_fields]:
                        result[particular_field_name] = [
                            result[particular_field_name], stored_values[particular_field_name]]
                else:
                    for particular_field_name in [*compound_key_values, *additional_fields]:
                        result[particular_field_name].append(
                            stored_values[particular_field_name])

        return result, result_count

    def _request_explicit_values(self, entity_class, key_fields, additional_fields, value):
        compound_values, compound_values_count = \
            self._compound_key_values(
                entity_class, key_fields, additional_fields, value, self._default_procedure_to_missed_key)
        # create relations
        if compound_values_count > 0:
            result = list(entity_class.objects.filter(
                **{k: v for k, v in compound_values.items()})) if compound_values_count == 1 else \
                list(entity_class.objects.filter(
                    **{'%s__in' % k: v for k, v in compound_values.items()}))
            result_index = {
                tuple([getattr(x, attname).pk for attname in key_fields]): x for x in result}
            # merge
            for i in range(0, compound_values_count):
                compound_key_value = tuple(
                    [v[i] if compound_values_count > 1 else v for k, v in compound_values.items() if k in key_fields])
                if not compound_key_value in result_index:
                    entity_instance = entity_class(
                        **{**{f'{key_fields[j]}_id': compound_key_value[j] for j in range(0, len(key_fields))},
                           **{additional_fields[j]: compound_values[additional_fields[j]][i] if compound_values_count > 1 else compound_values[additional_fields[j]] for j in range(0, len(additional_fields))}})
                    result_index[compound_key_value] = entity_instance
                else:
                    for j in range(0, len(additional_fields)):
                        setattr(result_index[compound_key_value],
                                additional_fields[j],
                                compound_values[additional_fields[j]][i] if compound_values_count > 1 else compound_values[additional_fields[j]])

            return result_index.values()
        else:
            return self.queryset.none()

    def _request_implicit_values(self, entity_class, value):
        key_fields = ('id', )
        compound_values, compound_values_count = \
            self._compound_key_values(
                entity_class, key_fields, (), value, self._default_procedure_to_missed_key)
        # create relations
        if compound_values_count == 1:
            result = list(entity_class.objects.filter(
                **{k: v for k, v in compound_values.items()}))
            return result if result else [entity_class(**{f'{k}' if k in key_fields else k: v for k, v in compound_values.items()})]
        elif compound_values_count > 1:
            result = list(entity_class.objects.filter(
                **{'%s__in' % k: v for k, v in compound_values.items()}))
            result_index = {
                tuple([getattr(x, attname) for attname in key_fields]): x for x in result}
            # merge
            for i in range(0, compound_values_count):
                compound_key_value = tuple(
                    [v[i] for k, v in compound_values.items()])
                if not compound_key_value in result_index:
                    result.append(entity_class(
                        **{**{f'{key_fields[i]}': compound_key_value[i] for i in range(0, len(key_fields))}}))
            return result
        else:
            return self.queryset.none()
        
    def bound_data(self, data, initial):
        """
        Return the value that should be shown for this field on render of a
        bound form, given the submitted POST data for the field and the initial
        data, if any.

        For most fields, this will simply be data; FileFields need to handle it
        a bit differently.
        """
        # if self.disabled:
        #    return initial
        # return data
        # Initial always actual @see clean method implementation
        return initial

    def clean(self, value):
        if self.widget and self.widget.field and self.widget.field.through:
            # Search by field. Eager approach
            through = self.widget.field.through
            compound_key = self.widget.field.key_data_components
            additional_fields = self.widget.field.additional_fields

            return self._request_explicit_values(through, compound_key, additional_fields, value)

        return self._request_implicit_values(self.widget.field.remote_field.model, value)



from django.apps import AppConfig as DjangoAppConfig

from django import VERSION
if VERSION >= (2, 0):
    from django.utils.translation import gettext_lazy as _
else:
    from django.utils.translation import ugettext_lazy as _

class AppConfig(DjangoAppConfig):
    """Configuration for the djeu app"""
    label = name = 'djeu'
    verbose_name = _('DjEu')



from django.db.models import fields
from djeu import forms


DEFAULT_HELP_TEXT = ' <b> Формат ввода даты: дд.мм.гг.: 24.04.1934. ' + \
    'При отсутствии параметра использовать 2 нижних подчеркивания: 24.__.1934 </b>'


class PartialDateField(fields.CharField):

    def __init__(self, *args, db_collation=None, max_length=64, help_text=DEFAULT_HELP_TEXT, **kwargs):
        super(PartialDateField, self).__init__(
            *args,
            db_collation=db_collation,
            max_length=max_length,
            help_text=help_text,
            **kwargs)

    def formfield(self, **kwargs):
        result = fields.Field.formfield(self, **{
            'form_class': forms.PartialDateField,
            **kwargs,
        })
        return result

    def value_to_string(self, obj):
        return super(PartialDateField, self).value_to_string(obj)

    def get_db_prep_value(self, value, connection, prepared=False):
        return super(PartialDateField, self).get_db_prep_value(value, connection, prepared)

    def to_python(self, value):
        if isinstance(value, str) or value is None:
            return value
        return str(value)

    def get_prep_value(self, value):
        value = super().get_prep_value(value)
        return self.to_python(value)

    def __str__(self):
        return super(PartialDateField, self).__str__()

    def __repr__(self):
        return self.__str__()
